#!/usr/bin/env python
"""Module pour le patron Registre.

Ce patron permet de garder un registre de toutes les instances des sous-classes
d'une classe donnée.

.. usage ::

    class RegistreElements(metaclass=MetaInstanceRegistre):
        pass

    class Element(RegistreElements):
        pass


Une instance de Element va être enregistrée dans RegistreElements.__instances
que l'on peut parcourir grâce à RegistreElements._instances(inclure_enfants=True).

NB: On peut également utiliser Element._instances() pour avoir accès à toutes les
instances d'Element.

Il s'agit d'un set que l'on peut ensuite parcourir pour trouver celle que l'on
souhaite.

NB: La classe fille doit être hashable. Si l'on utilise attrs il faut penser à déclarer
    @attr.s(eq=False) pour activer le hashage.
"""

import weakref


class MetaClasseRegistre(type):
    """Metaclass fournissant un registre de **classes**.

    inspiré de https://stackoverflow.com/a/48328119

    NB: non testé.
    """

    def __init__(cls, name, bases, attrs):
        # Create class
        super(MetaClasseRegistre, cls).__init__(name, bases, attrs)

        # Initialize fresh instance storage
        cls._classes = weakref.WeakValueDictionary()

    def __new__(cls, name, bases, attrs):
        new_cls = type.__new__(cls, name, bases, attrs)
        """
            Here the name of the class is used as key but it could be any class
            parameter.
        """

        # Store weak reference to class. WeakValueDictionnary will automatically remove
        # references to objects that have been garbage collected
        cls._classes[new_cls.__name__] = new_cls
        return new_cls

    def _recup_classes(cls, inclure_enfants=False):
        """Get all instances of this class in the registry.

        Si inclure_enfants=True, on renvoie aussi les instances des sous-classes.
        """
        classes = [] + list(cls._classes)[cls.__name__]
        if inclure_enfants:
            for child in cls.__subclasses__():
                classes += child._recup_classes(inclure_enfants=inclure_enfants)

        # Remove duplicates from multiple inheritance.
        return list(set(classes))


class MetaInstanceRegistre(type):
    """Metaclass fournissant un registre **d'instances**.

    source: https://stackoverflow.com/a/48328119

    Ici on utilise un "weakRefSet", donc on garde le registre de toutes les instances
    "en cours d'utilisation", ie dont on a encore la référence.
    Si on écrase la référence ou que l'on n'en a pas, alors l'instance disparaît du
    registre lors du passage du ramasse-miettes.

    NB: on peut préférer la composition (au lieu de l'héritage) dans ce genre de cas
        mais cette façon de faire me semble très "pythonesque" malgré tout.
    """

    def __init__(cls, name, bases, attrs):
        # Lors de la création **de la classe** du registre (et pas de l'instance)
        super(MetaInstanceRegistre, cls).__init__(name, bases, attrs)

        # Initialise le stockage des instances, pas besoin pour l'instant d'une "weak ref"
        cls.__instances = weakref.WeakSet()

    def __call__(cls, *args, **kwargs):
        # Crée l'instance (appelle les méthodes __init__ et __new__)
        inst = super(MetaInstanceRegistre, cls).__call__(*args, **kwargs)

        cls.__instances.add(inst)  # inst doit être hashable

        return inst

    def _instances(cls, inclure_enfants=False):
        """Renvoie les instances de **cette classe** stockées dans le registre.

        Si inclure_enfants=True, on renvoie aussi les instances des sous-classes.
        """
        instances = list(cls.__instances)
        if inclure_enfants:
            for child in cls.__subclasses__():
                instances += child._instances(inclure_enfants=inclure_enfants)

        # Remove duplicates from multiple inheritance.
        return list(set(instances))


class MetaInstancePersistanteRegistre(type):
    """Metaclass fournissant un registre **d'instances**.

    source: https://stackoverflow.com/a/48328119

    exemple d'implémentation dans automatheque.plugin

    Ici on utilise un "set" et non un "weakRefSet", pour le cas où on ne garde **pas**
    les références des instances crées : on les stocke juste dans le registre et on les
    cherche ensuite.
    Avec un weakRefSet on garde le registre de toutes les instances "en cours d'utilisation",
    (ie dont on a encore la référence).

    >>> class A(metaclass=MetaInstancePersistanteRegistre):
    >>>   pass
    >>> class B(metaclass=MetaInstanceRegistre):
    >>>   pass
    >>> a = A()
    >>> b = B()
    >>> A() # out: <__main__.A object at 0x7fcbc181b810>
    >>> B() # out: <__main__.B object at 0x7fcbc181a390>
    >>> A._instances()  # les 2 instances
    [<__main__.A object at 0x7fcbc181b810>, <__main__.A object at 0x7fcbc1a62d50>]
    >>> B._instances()  # une seule instance
    [<__main__.B object at 0x7fcbc181a3d0>]


    NB: on peut préférer la composition (au lieu de l'héritage) dans ce genre de cas
        mais cette façon de faire me semble très "pythonesque" malgré tout.
    """

    def __init__(cls, name, bases, attrs):
        # Lors de la création **de la classe** du registre (et pas de l'instance)
        super(MetaInstancePersistanteRegistre, cls).__init__(name, bases, attrs)

        # Initialise le stockage des instances, pas besoin pour l'instant d'une "weak ref"
        cls.__instances = set()

    def __call__(cls, *args, **kwargs):
        # Crée l'instance (appelle les méthodes __init__ et __new__)
        inst = super(MetaInstancePersistanteRegistre, cls).__call__(*args, **kwargs)

        cls.__instances.add(inst)  # inst doit être hashable

        return inst

    def _instances(cls, inclure_enfants=False):
        """Renvoie les instances de **cette classe** stockées dans le registre.

        Si inclure_enfants=True, on renvoie aussi les instances des sous-classes.
        """
        instances = list(cls.__instances)
        if inclure_enfants:
            for child in cls.__subclasses__():
                instances += child._instances(inclure_enfants=inclure_enfants)

        # Remove duplicates from multiple inheritance.
        return list(set(instances))



# -*- coding: utf-8 -*-
"""Liste des exceptions spécifiques à automatheque."""


class AutomathequeBaseException(Exception):
    """Classe de base dont héritent les autres."""


class DependanceManquante(AutomathequeBaseException):
    """Exception levée si une dépendance est manquante."""

    def __init__(self, dependance, msg=""):
        """Initialisation."""
        self.msg = "Dépendance {} manquante. {}.".format(dependance, msg)


class ArgumentManquant(AutomathequeBaseException):
    """Exception si un argument est manquant."""

    def __init__(self, argument):
        """Initialisation."""
        self.msg = "Argument '{}' manquant.".format(argument)



from .adaptateur import Adaptateur
from .borg import Borg
from .fabrique import Fabrique
from .monteur import Monteur
from .registre import (
    MetaInstanceRegistre,
    MetaInstancePersistanteRegistre,
    MetaClasseRegistre,
)
from .singleton import Singleton


__all__ = [
    "Adaptateur",
    "Borg",
    "Fabrique",
    "MetaClasseRegistre",
    "MetaInstanceRegistre",
    "MetaInstancePersistanteRegistre",
    "Monteur",
    "Singleton",
]



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Utilitaires de manipulation des répertoires

Divers utilitaires.
"""

import os
from pathlib import Path
import errno


from automatheque.log import recup_logger

LOGGER = recup_logger(__name__)


def mkdir_p(path):
    """Crée un répertoire et son arborescence : équivalent à  "mkdir -p".

    Renvoie une erreur si la cible existe et n'est pas un répertoire.

    TODO déprécier pour Path.mkdir()
    """
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


# Fonction supplémentaire à ajouter :
# TODO https://peterlyons.com/problog/2009/04/zip-dir-python


def supprime(repertoire, force=False):
    """Supprime un répertoire s'il est vide."""
    try:
        os.rmdir(repertoire)
        LOGGER.debug("Repertoire {} supprime car vide".format(repertoire))
    except OSError as ex:
        if ex.errno == errno.ENOTEMPTY and not force:
            LOGGER.warning("Repertoire {} non supprime car non vide".format(repertoire))


def remonte_arborescence(fichier, racine=None):
    """Générateur qui renvoie le niveau précédent de l'arborescence.

    Il s'arrête une fois que la racine est trouvée.

    :raise: ValueError si fichier n'est pas rattaché à la racine donnée
    """
    if not isinstance(fichier, Path):
        fichier = Path(fichier)
    racine = fichier.root if not racine else racine
    if not isinstance(racine, Path):
        racine = Path(racine)

    if fichier.relative_to(racine):
        parents_relatifs = [i for i in fichier.parents if i not in racine.parents]
        for parent in parents_relatifs[:-1]:
            yield parent


def iter_fichiers(source):
    """Itérateur de la liste des fichiers d'une source Path.

    On crée un itérateur que l'on ait reçu un fichier ou un rep en entrée !

    :param source: Source des fichiers : 1 rep ou 1 fichier
    :type source: Path
    """
    if source.is_dir():
        # On récupère tous les fichiers récursivement :
        fichiers = source.rglob("*")
    else:
        # On ne récupère que le fichier d'entrée :
        fichiers = source.parent.glob(source.name)
    return [f for f in fichiers if not f.is_dir()]



import abc


class StockageAbstraite(abc.ABC):
    @abc.abstractmethod
    def existe(self, reference: str) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def sauvegarde(self, reference: str, contenu: str) -> bool:
        raise NotImplementedError



# -*- coding=utf-8 -*-
"""Gestion des Greffons et de leur capacités.
"""
import uuid

import attr

from automatheque.configuration import charge_configuration
from automatheque.greffon.registre import RegistreGreffons
from automatheque.log import recup_logger
from automatheque.util.classe import classproperty


LOGGER = recup_logger(__name__)
LOGGER.debug("LOGGER")


@attr.s(eq=False)  # pour rendre la classe hashable pour le set du registre
class Greffon(RegistreGreffons):
    """
    https://fr.wikipedia.org/wiki/Plugin : on utilise "greffon" au lieu de "module d'extension" car
    c'est plus compact et moins prône à la confusion, malgré la recommandation de la DGLFLF.

    ## Pourquoi utiliser des greffons :

    * "Donner accès, pour une même fonction, à différentes solutions" :
        * = pouvoir modifier le comportement de l'application, notamment en fonction de la conf
        * ex: pour récupérer les données nutritionnelles d'un aliment, je peux utiliser l'api API1
        grâce au greffon adéquat, ou l'api API2 via son greffon, et je décide en configuration
        laquelle il faut utiliser et quels arguments pour chaque greffon (login etc.), les 2
        greffons partageant les mêmes "capacités"
    * Peut évoluer plus vite, car il est plus simple de créer un nouveau greffon pour une nouvelle
    interface (pour une nouvelle api par exemple, ou un nouveau site de données, ou un nouveau
    traitement) que d'attendre une nouvelle version du logiciel.

    ## Quel greffon récupérer :

    * un par un, par le biais du nom du greffon
        * si on veut un traitement spécifique, ex: je veux utiliser l'api openfoodfacts, ou bien je
          veux parser un fichier qif vs un fichier csv ...)
    * tous ceux qui rendent un service donné :
        * on exécute le premier qui marche
            * ex: j'utilise la capacité "ALIMENT.PEUPLER_INFOS_NUTRITIONNELLES", et je m'arrête
              après le premier appel OK
        * ou un accumule les données (idem mais je complète les données à chaque appel d'un nouveau
          greffon)
        * NB le choix entre ces 2 manières de fonctionner dépend du logiciel qui utilise les
          greffons, ici automatheque se contente de les rendre accessibles et recherchables.
    """

    # Utilisation de kw_only=True pour l'héritage :
    # https://github.com/python-attrs/attrs/issues/38
    # Il faut utiliser l'option également dans les classes filles.
    identifiant = attr.ib(default=str(uuid.uuid4()), kw_only=True)
    config_requise = attr.ib(default=False, init=False, kw_only=True)
    config = attr.ib(init=False, factory=charge_configuration, kw_only=True)

    @classproperty
    def cle(cls):
        """Retourne le nom de la classe sans "Greffon" s'il existe, en minuscule"""
        return cls.__name__.lower().replace("greffon", "")

    @property
    def capacites(self):
        return self.CAPACITES

    @property
    def actif(self):
        """TODO pas très clair le fonctionnement !"""
        if not self.config_requise or self.config:
            return True
        return False

    def _signale_appel(self):
        # TODO log !!!
        # TODO on devrait mettre une annotation sur tous les appels de type "capacites"
        LOGGER.debug("Greffon : {} - debut appel".format(self.cle))



"""Package comprenant quelques outils classiques."""

# Imports depuis fichier
from .fichier import enleve_caracteres_invalides

# Imports depuis repertoire
from .repertoire import mkdir_p

# Imports depuis structures_python
from .structures_python import dict_merge

__all__ = [
    # .fichier
    "enleve_caracteres_invalides",
    # .repertoire
    "mkdir_p",
    # .structures_python
    "dict_merge",
]



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Utilitaires de manipulation des structures classiques de python.

Divers utilitaires pour dict, list etc.
"""
from copy import deepcopy
from datetime import date, datetime, time

import pytz


def dict_merge(a, b):
    """Merge récursif de 2 dictionnaires python.

    https://www.xormedia.com/recursively-merge-dictionaries-in-python/

    recursively merges dict's. not just simple a['key'] = b['key'], if
    both a and b have a key who's value is a dict then dict_merge is called
    on both values and the result stored in the returned dictionary.
    """
    if not isinstance(b, dict):
        return b
    result = deepcopy(a)
    for k, v in b.items():
        if k in result and isinstance(result[k], dict):
            result[k] = dict_merge(result[k], v)
        else:
            result[k] = deepcopy(v)
    return result


def date_en_datetime(
    date_src, time_obj=None, heures=0, minutes=0, secondes=0, tzinfo=pytz.utc
):
    """Parfois on reçoit une date et on veut la transformer en datetime.

    Si on ne précise rien, elle est définie pour la timezone UTC.

    :param date_src: date à transformer éventuellement en datetime (si elle ne
                     l'est pas déjà)
    :param time_obj: Objet datetime.time à utiliser pour compléter la date
                     defaut = None
    :param heures: si on n'envoie pas d'objet "time_obj" on peut donner ses
                   paramètres
    :param minutes: idem
    :param secondes: idem
    :param tzinfo: Objet pytz timezone, defaut à pytz.utc
    """
    date_conv = date_src
    if time_obj is None:
        time_obj = time(heures, minutes, secondes, tzinfo=tzinfo)
    elif time_obj is not None and date_est_naive(time_obj):
        tzinfo.localize(time_obj)
    if not isinstance(date_src, datetime) and isinstance(date_src, date):
        date_conv = datetime.combine(date_conv, time_obj)
    return date_conv


def date_est_naive(date_src):
    """Renvoie True si la date est naive (ie sans TZ), False sinon."""
    if date_src.tzinfo is not None and date_src.tzinfo.utcoffset(date_src) is not None:
        return False
    return True


def datetimezone_depuis_code_pays(date_src, code_pays):
    """Renvoie une datetime "aware" à partir d'un code pays.

    Si la date est déjà "aware", ne fait rien. Si la date est bien "naive" en
    revanche, on cherche la première timezone associée au code pays, et on
    l'ajoute à la datetime.

    :param date_src: datetime "naive" (sinon retourne date_src)
    :param code_pays: code du pays (ex: FR, JP)
    :return: datetime ou None si le code_pays est invalide
    """
    if not date_est_naive(date_src):
        return date_src

    try:
        codes = pytz.country_timezones[code_pays]
    except KeyError:
        return None
    else:
        return pytz.timezone(codes[0]).localize(date_src)



from copy import deepcopy

from automatheque.conception.structures import Fabrique, Monteur
from automatheque.configuration import charge_configuration, ConfigParser
from automatheque.greffon.greffon import Greffon
from automatheque.log import recup_logger


LOGGER = recup_logger(__name__)


class FabriqueGreffon(Fabrique):
    """Classe qui permet d'instancier les Greffons dont les Monteur ont été enregistrés.

    Il faut dans un premier temps enregistrer les monteurs des greffons par la fonction
    enregistre_greffons, puis il faut activer le greffon donné ce qui lance
    `monteur.cree()` avec les arguments donnés.

    À défaut de Monteur, on peut donner la classe qui sera instanciée directement.
    Dans les deux cas il faut que la classe ait une propriété "clé".

    Si on utilise un Monteur, on peut utiliser la même clé que le Greffon, tant qu'elle
    n'est pas utilisée par plusieurs Greffons ou Monteurs, puisqu'elle servira à
    déclencher l'instanciation du Monteur ou du Greffon associée.

    .. example::
        >>> from automatheque.greffon import fabrique_greffon
        >>> fabrique_greffon.charge_monteurs([
               "plugin1.monteur.Monteur1",
               "plugin2.monteur.Monteur2"
            ])
        >>> plugin1_id1 = fabrique_greffon.active(
               "monteur1",
               *args,
               identifiant="plugin1_identifiant1",
               **kwargs
            )
        >>> plugin1_id1.appel_quelconque_capacite()
    """

    def active(self, cle_monteur, *args, identifiant=None, **kwargs):
        """Va chercher dans le registre des greffons et renvoie le greffon activé.

        S'il ne trouve pas le greffon il en instancie un autre.
        TODO : comme identifiant peut etre auto génré on devrait pouvoir
        le rendre facultatif si on ne veut pas l'appeler nommément ? ou mettre clé ?
        """
        if not identifiant or identifiant not in Greffon.greffons_identifiants():
            # greffons.keys():
            try:
                if identifiant:
                    kwargs["identifiant"] = identifiant
                instance_greffon = self.cree(cle_monteur, *args, **kwargs)
                LOGGER.debug(f"Greffon activé : {instance_greffon}")
                if not instance_greffon.actif:
                    LOGGER.warning(f"Greffon {instance_greffon} inactif")
                    return False
            except Exception:
                LOGGER.exception(f"Echec activation greffon : {identifiant}")
        return Greffon.greffon_par_identifiant(identifiant)

    def charge_monteurs(self, liste_monteurs):
        """Enregistre les monteurs de greffons disponibles dans la fabrique.

        Cela  pourrait etre fait de manière automatique, par ex en scannant un
        répertoire.
        Dans tous les cas ensuite il faut les activer un par un dans la configuration.
        """
        monteurs = []
        for elem in liste_monteurs:
            from pydoc import locate

            monteur = locate(elem)  # ou importlib TODO ?
            if issubclass(monteur, Greffon):
                monteur_concret = monteur
            elif issubclass(monteur, Monteur):
                monteur_concret = monteur()  # Il faut instancier le Monteur
            else:
                raise ValueError(
                    "La fabrique ne peut charger que des Greffons ou des Monteurs"
                )
            monteurs.append(self.enregistre_monteur(monteur.cle, monteur_concret))
        return monteurs

    # def greffons_charge_defaut(self):
    #    """Charge les greffons par défaut dans la liste de monteurs."""
    #    self.enregistre_greffons(LISTE_GREFFONS_DEFAUT)
    #    LOGGER.debug(f"Greffons chargés : {self.recup_monteurs()}")

    def active_greffons_conf(self, liste_greffons=None, configuration=None):
        """Active les greffons qui sont définis en conf dans l'option "greffons".

        Si on utilise cette méthode pour activer les greffons, alors il est nécessaire
        de fournir un identifiant, on ne peut pas laisser automatheque créer l'identifiant
        automatiquement.

        .. example::
            [greffons]
            # liste d'identifiants des greffons, il peut y avoir plusieurs identifiants
            # pour le même greffon, par exemple avec des arguments différents
            greffons=kodi1,trakt1

            [kodi1]
            # clé du monteur ou du plugin
            greffon=kodi
            # Argument supplémentaire fourni à l'instanciation
            hote=xxx
            [trakt1]
            greffon=trakt

        configuration peut aussi être un dictionnaire équivalent.
        """
        if configuration is None:
            configuration = dict(charge_configuration())
        elif isinstance(configuration, ConfigParser):
            configuration = dict(configuration)
        elif not isinstance(configuration, dict):
            raise ValueError("'configuration' n'est pas au bon format")

        if liste_greffons is None:
            try:
                # On laisse dans "greffon" si on veut sortir le code un jour !
                liste_greffons = configuration.get("greffons").get("greffons")
                liste_greffons = [i.strip() for i in liste_greffons.split(",")]
            except Exception as e:
                LOGGER.exception(e)
                raise ValueError(f"pas de liste de greffons dans {configuration}")

        greffons_actifs = []
        for identifiant in liste_greffons:
            conf = deepcopy(configuration.get(identifiant))
            cle = conf.pop("greffon")
            greffons_actifs.append(self.active(cle, identifiant=identifiant, **conf))
        return greffons_actifs


fabrique_greffon = FabriqueGreffon()



# -*- coding: utf-8 -*-
"""Module pour le patron Borg.

Spécifique à python, ce patron a le même but que le patron Singleton, mais au lieu de
forcer l'existence d'une unique instance d'une classe, le patron Borg permet d'avoir
plusieurs instances mais qui partagent le même état.

[wikipedia](https://fr.wikipedia.org/wiki/Singleton_(patron_de_conception)#Consid%C3%A9rations_avanc%C3%A9es)

source: https://github.com/faif/python-patterns/blob/master/patterns/creational/borg.py
"""


class Borg:
    """Classe "Singleton-like" pour partager les états de plusieurs instances."""

    # variable de classe contenant l'état à partager
    __etat_partage = {}

    def __init__(self):
        # copie de l'état lors de l'initialisation d'une nouvelle instance
        self.__dict__ = self.__etat_partage




# -*- coding: utf-8 -*-
"""Module pour le patron Singleton.

Ce patron vise à assurer qu'il n'y a toujours qu'une seule instance d'une classe en
fournissant une interface pour la manipuler. C'est un des patrons les plus simples.

L'objet qui ne doit exister qu'en une seule instance comporte une méthode pour obtenir
cette unique instance et un mécanisme pour empêcher la création d'autres instances.
Mais en python ce n'est pas nécessaire, la variable de classe "cls.instance" suffit.

[wikipedia](https://fr.wikipedia.org/wiki/Singleton_(patron_de_conception))

NB: ! attention il s'agit d'un "antipattern"
"""


class Singleton(object):
    """Classe Singleton qui permet l'héritage."""

    # Dictionnaire Python référençant les instances déjà créés : une pour chaque
    # sous-classe.
    _singletons = {}

    def __new__(cls, *args, **kargs):
        if Singleton._singletons.get(cls) is None:
            Singleton._singletons[cls] = object.__new__(cls, *args, **kargs)
        return Singleton._singletons[cls]



"""Module d'aide à la gestion des dépendances externes.

Il est notamment utilisé :
* dans le package "dependance", qui sert à livrer des dépendances complètes avec automatheque
  (comme pour pyexiftool) et/ou à vérifier leur installation localement (en attendant peut être
  l'utilisation de poetry ou autre ?).
* dans des scripts qui utilisent automatheque, pour les aider à vérifier la présence de dépendances
  externes

# Fonctionnement :

Ce module fournit un registre des dépendances chargées.

Lorsqu'une dépendance est chargée, une instance de la classe Executant est créée, et stockée dans
le registre.

Ainsi c'est ce registre qu'il faut appeler si on a besoin d'une dépendance.

Il est ensuite conseillé de créer ses propres classes "wrapper" et de leur donner l'instance de
l'exécutant en paramètre, qui se chargera de faire les appels subprocess.

"""
from collections import namedtuple
import os
import shutil
import subprocess

import attr

from automatheque.exceptions import DependanceManquante
from automatheque.log import recup_logger


LOGGER = recup_logger(__name__)


MAUVAISE_CLE_ERREUR = """Aucune dépendance à ce nom n'a été trouvée.
Vérifiez que vous avez bien chargé la dépendance précédemment."""
registre_dependances = {}
Dependance = namedtuple("Dependance", ["erreur", "presente", "executant"])


def verifie_dependance(cle_dependance):
    """Vérifie que la dépendance externe donnée est bien chargée.

    Renvoie une erreur si la dépendance n'est pas chargée ou que l'exécutable n'est pas présent,
    et ne renvoie rien sinon.

    :returns: None
    """
    try:
        d = registre_dependances[cle_dependance]
        if not d.presente:
            raise DependanceManquante(cle_dependance, d.erreur)
    except KeyError:
        raise DependanceManquante(cle_dependance, MAUVAISE_CLE_ERREUR)


def charge_dependance(cle, executable, executable_complet, erreur="", verifie=True):
    """Charge la dépendance donnée, dans le registre.

    TODO: voir s'il faut utiliser une fabrique et/ou un constructeur ...
    """
    executable = recup_executable(executable, executable_complet)
    executant = Executant(executable)

    # Ajout dans le registre des dépendances :
    d = {
        "erreur": erreur if erreur else _gen_erreur(cle),
        "presente": executable,
        "executant": executant,
    }
    registre_dependances[cle] = Dependance(**d)

    # On verifie que la dépendance existe et est bien chargée :
    if verifie:
        verifie_dependance(cle)

    # Si on arrive là alors la dépendance est bien vérifiée, on retourne "executant"
    return executant


def recup_executable(nom, chemin_complet):
    """Récupère le chemin vers l'exécutable demandé.

    :returns: str ou False
    """
    path = shutil.which(nom)
    # TODO : il faut aussi tester l'execution sans args pour voir si ç march non ?
    # Si on ne trouve pas l'exécutable on essaie de forcer un chemin classique :
    if path is None:
        path = chemin_complet
        if not os.path.isfile(path) or not os.access(path, os.X_OK):
            return False
    return path


def _gen_erreur(cle):
    """Génère une erreur standard pour la dépendance donnée."""
    return "\n{} n'est pas installé !\n".format(cle.title()).lstrip()


# TODO
"""
def verifie_specifique(Dependance d):
    for f in d.fichiers_specifiques:
        if not os.path.exists(f):
            raise DependanceIncomplete(d["xxx"])
"""


@attr.s
class Executant(object):
    """Classe pour faciliter l'exécution des dépendances."""

    executable = attr.ib()

    def exec(self, *args, stdin=None, **kwargs) -> subprocess.CompletedProcess:
        """Execute le programme donné en dépendances avec les arguments en paramètres.

        Renvoie un objet CompletedProcess (https://docs.python.org/3/library/subprocess.html#subprocess.CompletedProcess)

        L'idéal est de donner une instance de cette classe à un wrapper pour l'action souhaitée
        qui se charge d'appeler `self.executant.exec` par ex.
        """
        procargs = [self.executable, *args]
        for k, v in kwargs.items():
            procargs += [k, v]

        msg_redir = " < 'stdin'" if stdin is not None else ""
        LOGGER.debug(f"Executant.exec : procargs={procargs}" + msg_redir )
        pipes = subprocess.run(
            procargs,
            stdin=stdin if stdin is not None else subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        return pipes
